# Gtk-Stream : A stream-based GUI protocol
# Copyright (C) 2024  Marc Coiffier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import threading
import signal

from . import Gtk, GLib, Gdk
from .common import print_event
from .properties import parse_property, get_prop_type, set_parse_prop

class _Object:
    pass

def app_message(name, store = None):
    """A decorator for methods that are both called from the pilot
    application and need access to the gtk main thread"""
    def app_message_f(f):
        def ret(self, *args, **kwargs):
            def cb():
                f(self, *args, **kwargs)
            self.run_when_idle(cb)
        ret.__tag_name__ = name
        ret.__store__ = store
        return ret
    return app_message_f

def single_store():
    store = _Object()
    def setChild(child):
        store.child = child
    return (lambda: store.child, setChild, None)
def multiple_store():
    children = []
    return (lambda: children, children.append, None)
def style_store():
    style = []
    return (lambda: " ".join(style),None, style.append)

class GtkStreamApp(Gtk.Application):
    def __init__(self, logger, **kwargs):
        super().__init__(**kwargs)
        self.logger = logger
        self.named_widgets = { }
        self.named_windows = { }
        
        # The first messages from the pilot may arrive before the
        # application is ready to process them.
        # 
        # If that happens, store the actions until they can be taken
        # (when the "startup" signal is called)
        callback_queue = []
        def run_when_idle_before_startup(cb):
            callback_queue.append(cb)
        self.run_when_idle = run_when_idle_before_startup
        
        def on_startup(_):
            for cb in callback_queue:
                GLib.idle_add(cb)
            self.run_when_idle = GLib.idle_add

            def on_sigint(a,b):
                self.logger.info("SIGINT received, quitting application")
                self.quit()
            signal.signal(signal.SIGINT, on_sigint)
        self.connect('startup', on_startup)

        def on_activate(a):
            a.hold()
        self.connect('activate', on_activate)

        self.attrs_set = threading.Semaphore()
        self.attrs_set.acquire()
        
    def name_widget(self, id, w):
        if id is not None:
            self.named_widgets[id] = w

    def set_attrs(self, attrs):
        for name, val in attrs.items():
            set_parse_prop(self, self, name, val)
        self.attrs_set.release()
            
    @app_message('file-dialog')
    def open_file_dialog(self, id, parent):
        dialog = Gtk.FileDialog()
        dialog.props.modal = True
        def on_choose(_, b):
            try:
                file = dialog.open_finish(b)
                print(f"{id}:selected:{file.get_path()}")
                sys.stdout.flush()
            except GLib.GError as e:
                print(f"{id}:none-selected")
                sys.stdout.flush()
                
        dialog.open(parent = self.named_windows[parent], callback = on_choose)

    @app_message('window', single_store)
    def new_window(self, document, id, **attrs):
        win = Gtk.Window(application=self)
        for (attr_name, attr_val) in attrs.items():
            self.logger.debug("Setting attr '%s' on window", attr_name)
            set_parse_prop(self, win, attr_name, attr_val)
        self.named_windows[id] = win
        win.set_child(document.render())
        win.connect('close-request', print_event(self.logger, 'close-request', id))
        win.present()

    @app_message('style', style_store)
    def add_style(self, style):
        provider = Gtk.CssProvider()
        provider.load_from_data(style)
        Gtk.StyleContext.add_provider_for_display(Gdk.Display.get_default(), provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    @app_message('add-icon-path')
    def add_icon_path(self, path):
        theme = Gtk.IconTheme.get_for_display(Gdk.Display.get_default())
        theme.add_search_path(path)
        
    @app_message('close-window')
    def close_window(self, id):
        self.named_windows[id].close()

    @app_message('remove')
    def remove_widget(self, id):
        w = self.named_widgets[id]
        w.get_parent().remove(w)

    @app_message('insert', multiple_store)
    def insert_widgets(self, documents, into):
        for doc in documents:
            self.insert_widget(doc, into)

    def insert_widget(self, document, into):
        if into in self.named_widgets:
            w = self.named_widgets[into]
            w.insert_child(document)
        else:
            raise Exception(f"Error: unknown widget id '{into}'")
    
    @app_message('set-prop')
    def set_prop(self, id, name, value):
        w = self.named_widgets[id]
        w.set_property(name, parse_property(get_prop_type(w.__class__, name), value)(self))




# Gtk-Stream : A stream-based GUI protocol
# Copyright (C) 2024  Marc Coiffier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ... import Gtk
from .. import Document

class Paned(Document):
    __g_class__ = Gtk.Paned
    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
    def render(self):
        if len(self.children) == 1:
            return self.children[0].render()
        
        l = self.children[0].render()
        for (r,rem_size) in zip(self.children[1:], range(len(self.children),1,-1)):
            j = Gtk.Paned()
            j.set_shrink_start_child(False)
            j.set_shrink_end_child(False)
            j.props.start_child = l
            j.props.end_child = r.render()
            self.set_properties(j)
            l = j
        return l



# Gtk-Stream : A stream-based GUI protocol
# Copyright (C) 2024  Marc Coiffier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ... import Gtk
from .. import Document, PseudoDocument

class FrameLabel(PseudoDocument):
    def __init__(self, app):
        super().__init__(app)

class Frame(Document):
    __g_class__ = Gtk.Frame
    def __init__(self, app, **kwargs):
        super().__init__(app,**kwargs)
    def render_raw(self):
        return Gtk.Frame()
    def insert_child(self, w, d):
        if isinstance(d, FrameLabel):
            w.set_label_widget(d.render())
        else:
            w.set_child(d.render())



# Gtk-Stream : A stream-based GUI protocol
# Copyright (C) 2024  Marc Coiffier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ... import Gtk, Gio, GObject
from .. import Document

class Item(Document):
    def __init__(self, app, value, **kwargs):
        super().__init__(app, **kwargs)
        self.child = None
        self.value = value
    def add_child(self, child):
        self.child = child
    def render(self):
        return self.child.render()
class _ItemObject(GObject.Object):
    def __init__(self, doc):
        super().__init__()
        self.doc = doc
        self.value = doc.value

    @GObject.Property(type=str)
    def item_value(self):
        return self.value
        
class Dropdown(Document):
    __g_class__ = Gtk.DropDown
    def __init__(self, app, id, **kwargs):
        super().__init__(app, id=id, **kwargs)
    def render_raw(self):
        model = Gio.ListStore(item_type=_ItemObject)
        for item in self.children:
            model.append(_ItemObject(item))

        factory = Gtk.SignalListItemFactory()
        def on_list_setup(_, list_item):
            list_item.item_value = None
        def on_list_bind(_, list_item):
            item = list_item.get_item()
            if list_item.item_value != item.value:
                list_item.item_value = item.value
                list_item.set_child(item.doc.render())
        factory.connect("setup", on_list_setup)
        factory.connect("bind", on_list_bind)

        ret = Gtk.DropDown(
            model=model,
            expression=Gtk.PropertyExpression.new(_ItemObject, None, 'item_value'),
            factory=factory)
        
        self.connect_event(
            ret, 'notify::selected-item', 'selected',
            get_data = lambda w,_: w.get_selected_item().value)
        return ret
    def insert_child(self, w, d):
        pass



# Gtk-Stream : A stream-based GUI protocol
# Copyright (C) 2024  Marc Coiffier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse

def main():
    parser = argparse.ArgumentParser(
        prog='gtk-stream',
        description='A stream-based GUI tool',
        add_help=False
    )
    parser.add_argument('--list-icons', action='store_true')
    parser.add_argument('--hook')
    parser.add_argument('-h', '--help', action='store_true')
    args = parser.parse_args()

    if args.help:
        parser.print_help()

    elif args.list_icons:
        from . import Gtk, Gdk
        theme = Gtk.IconTheme.get_for_display(Gdk.Display.get_default())
        for name in theme.get_icon_names():
            print(name)

    elif args.hook:
        match args.hook:
            case "bash":
                print('''
coproc GTK_STREAM { exec gtk-stream; }
trap 'kill -INT "$GTK_STREAM_PID"; wait "$GTK_STREAM_PID"' EXIT
exec {GTK_STREAM_EVENTS}<&"${GTK_STREAM[0]}"
exec {GTK_STREAM_MESSAGES}>&"${GTK_STREAM[1]}"
function GTK.send() {
    if (($# > 0)); then
        printf "$@" >&"$GTK_STREAM_MESSAGES"
    else
        cat >&"$GTK_STREAM_MESSAGES"
    fi
}
function GTK.receive() {
    read -u "$GTK_STREAM_EVENTS" "$@"
}
''')
    else:
        import io
        import sys
        import os
        import threading
        import xml.sax as sax
        
        from .parser import GtkStreamXMLHandler
        from .application import GtkStreamApp
        from .common import Logger, LogLevel

        class GtkStreamErrorHandler(sax.handler.ErrorHandler):
            def error(self, exc):
                raise exc
            def fatalError(self, exc):
                raise exc

        logLevel = LogLevel.__dict__.get(os.environ.get('GTK_STREAM_LOGLEVEL', 'WARN'), LogLevel.WARN)
        
        logger = Logger(logLevel)
        
        app = GtkStreamApp(logger)

        def parser_main():
            handler = GtkStreamXMLHandler(app)
            errHandler = GtkStreamErrorHandler()
            parser = sax.make_parser()
            parser.setContentHandler(handler)
            parser.setErrorHandler(errHandler)
            try:
                parser.parse(io.FileIO(0, 'r', closefd=False))
            except Exception as e:
                logger.error("Done with exception : %s", e)
            handler.quit_application()

        parser_thread = threading.Thread(target = parser_main, daemon = True)
        parser_thread.start()

        # When the parser parses the 'application' tag, it sets the
        # application attributes, and releases the semaphore so the
        # application can start
        app.attrs_set.acquire() 
        app.run()



# Gtk-Stream : A stream-based GUI protocol
# Copyright (C) 2024  Marc Coiffier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys

from . import Gtk, Gio

def _const(x):
    return lambda _: x
def _parse_orientation_property(val):
    return _const((Gtk.Orientation.HORIZONTAL) if val == "horizontal" else (Gtk.Orientation.VERTICAL))
def _parse_boolean_property(val):
    return _const(True if val == "true" else False)
def _parse_float_property(val):
    return _const(float(val))
def _parse_int_property(val):
    return _const(int(val))
def _parse_searchMode_property(val):
    match val:
        case 'exact':
            return _const(Gtk.StringFilterMatchMode.EXACT)
        case 'substring':
            return _const(Gtk.StringFilterMatchMode.SUBSTRING)
        case _:
            return _const(Gtk.StringFilterMatchMode.PREFIX)
def _parse_adjustment_property(val):
    adj = Gtk.Adjustment()
    start, end, *rest = val.split(':')
    adj.set_lower(int(start))
    adj.set_upper(int(end))
    if len(rest) > 0:
        default = rest[0]
    else:
        default = start
    adj.set_value(int(default))
    return _const(adj)
def _parse_css_classes_property(val):
    return _const(val.split())
def _parse_widget_property(val):
    return lambda app: app.named_widgets[val]
def _parse_window_property(val):
    return lambda app: app.named_windows[val]
def _parse_gfile_property(val):
    return _const(Gio.File.new_for_path(val))
def _parse_align_property(val):
    match val:
        case "start":
            return _const(Gtk.Align.START)
        case "center":
            return _const(Gtk.Align.CENTER)
        case "end":
            return _const(Gtk.Align.END)
def _parse_selection_mode_property(val):
    match val:
        case "none":
            return _const(Gtk.SelectionMode.NONE)
        case "single":
            return _const(Gtk.SelectionMode.SINGLE)
        case "browse":
            return _const(Gtk.SelectionMode.BROWSE)
        case "multiple":
            return _const(Gtk.SelectionMode.MULTIPLE)
        
_PARSE_TYPE_PROPERTY = {
    'GStrv'                    : _parse_css_classes_property,
    'GtkOrientation'           : _parse_orientation_property,
    'gdouble'                  : _parse_float_property,
    'gfloat'                   : _parse_float_property,
    'gint'                     : _parse_int_property,
    'gboolean'                 : _parse_boolean_property,
    'GtkStringFilterMatchMode' : _parse_searchMode_property,
    'GtkWidget'                : _parse_widget_property,
    'GtkStack'                 : _parse_widget_property,
    'GtkWindow'                : _parse_window_property,
    'GtkAdjustment'            : _parse_adjustment_property,
    'gchararray'               : _const,
    'GFile'                    : _parse_gfile_property,
    'GtkAlign'                 : _parse_align_property,
    'GtkSelectionMode'         : _parse_selection_mode_property,
}

def parse_property(prop_type, val):
    # print(f"Parsing property '{val}' of type '{prop_type}'", file=sys.stderr)
    return _PARSE_TYPE_PROPERTY[prop_type](val)
def get_prop_type(klass, prop):
    try:
        return klass.find_property(prop).value_type.name
    except AttributeError:
        raise Exception(f"Unknown GTK property '{prop}' of class '{klass}'")
def set_parse_prop(app, w, prop_name, val):
    w.set_property(prop_name, parse_property(get_prop_type(w.__class__, prop_name), val)(app))



# Gtk-Stream : A stream-based GUI protocol
# Copyright (C) 2024  Marc Coiffier
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import functools
import threading
import signal
import sys
import xml.sax as sax

from . import GLib
from . import documents as docs
from .properties import set_parse_prop

WIDGET_DOCUMENTS = {
    'progress-bar'    : docs.ProgressBar,
    'label'           : docs.Label,
    'box'             : docs.Box,
    'box-prepend'     : docs.BoxPrepend,
    'button'          : docs.Button,
    'dropdown'        : docs.Dropdown,
    'item'            : docs.Item,
    'paned'           : docs.Paned,
    'grid'            : docs.Grid,
    'cell'            : docs.Cell,
    'frame'           : docs.Frame,
    'frame-label'     : docs.FrameLabel,
    'link'            : docs.LinkButton,
    'switch'          : docs.Switch,
    'picture'         : docs.Picture,
    'icon'            : docs.Icon,
    'separator'       : docs.Separator,
    'scrolled-window' : docs.ScrolledWindow,
    'stack'           : docs.Stack,
    'flow-box'        : docs.FlowBox,
    'flow-box-prepend': docs.FlowBoxPrepend,
    'entry'           : docs.Entry,
    'scale'           : docs.Scale,
    'stack-side-bar'  : docs.StackSidebar,
    'stack-page'      : docs.StackPage,
}

class GtkStreamXMLHandler(sax.ContentHandler):
    def __init__(self, app):
        self.app = app
        self.logger = app.logger

        self.transition_enter = self.transE_conn
        self.transition_leave = self.transL_final
        self.transition_chars = self.ignore_chars

        # Get all messages directly from the application
        # class. This allows defining new messages without
        # touching the parser
        self.messages = {
            f.__tag_name__: self.start_message(functools.partial(f,self.app), f.__store__)
            for f in self.app.__class__.__dict__.values()
            if hasattr(f, '__tag_name__')
        }

    def quit_application(self):
        def cb():
            self.logger.info("Quitting app")
            self.app.quit()
        GLib.idle_add(cb)

    def ignore_chars(self, s):
        pass
        
    def transE_final(self, name, attrs):
        raise Exception(f"Unexpected tag '{name}'")
    def transL_final(self, name):
        raise Exception(f"Unexpected end tag '{name}'")
    def transL_tag(self, tag, enter, leave_parent, leave = None):
        def ret(name):
            if name == tag:
                if leave != None:
                    leave()
                self.transition_enter = enter
                self.transition_leave = leave_parent
            else:
                raise Exception(f"Error: expected end tag '{tag}', got '{name}'")
        return ret
                               
    def transE_conn(self, name, attrs):
        match name:
            case 'application':
                self.app.set_attrs(attrs)
                self.transition_enter = self.transE_message
                self.transition_leave = self.transL_tag('application', self.transE_final, self.transL_final)
            case _:
                raise Exception("Error: expected 'application' tag")

    def start_message(self, f, child = None):
        if child != None:
            def ret(name,attrs):
                getC, setC, setChars = child()
                old_chars = self.transition_chars
                def leave():
                    if setChars != None:
                        self.transition_chars = old_chars
                    f(getC(),**attrs)
                if setChars != None:
                    self.transition_chars = setChars
                self.transition_enter = self.transE_addChild(setC)
                self.transition_leave = self.transL_tag(name, self.transE_message, self.transition_leave, leave)
        else:
            def ret(name, attrs):
                def leave():
                    f(**attrs)
                self.transition_enter = self.transE_final
                self.transition_leave = self.transL_tag(name, self.transE_message, self.transition_leave, leave)
        return ret

    def transE_message(self, name, attrs):
        start = self.messages.get(name)
        if start != None:
            start(name,attrs)
        else:
            raise Exception(f"Error: unknown message '{name}'")

    def transE_addChild(self, addChild):
        def ret(name, attrs):
            leave_parent = self.transition_leave
            doc_class = WIDGET_DOCUMENTS.get(name)
            if doc_class != None:
                doc = doc_class(self.app, **attrs)
                addChild(doc)
                self.transition_enter = self.transE_addChild(lambda child: doc.add_child(child))
                self.transition_leave = self.transL_tag(name, self.transE_addChild(addChild), leave_parent)
            else:
                raise Exception(f"Error: Unknown widget type '{name}'")
        return ret

    def characters(self, s):
        self.transition_chars(s)
    def startElement(self, name, attrs):
        self.transition_enter(name, attrs)
    def endElement(self, name):
        self.transition_leave(name)
